#include <stdio.h>
#include <math.h>
#include "invsubs.h"


/*
 * Function to return a row of a "Laplacian" Smoothness constraint sub-matrix
 * for a 2D image
 * 
 * ASSUMES THAT THE IMAGE INDEXING BEGINS WITH ONE, SO THE FIRST ROW CONTAINS
 * ELEMENTS 1...NCOLS
 * 
 * the differencing star has:
 * 
 * 2(vwt+hwt)*smoother in the center
 * 
 * -vwt*smoother above/below
 * 
 * -hwt*smoother left/right
 * 
 * smoother = the smoothness weight value;
 * 
 * vwt = the relative vertical difference weight;
 * 
 * hwt = the relative horizontal weight;
 * 
 * whichParameter = the index of the "parameter" to be constrained (1...nr*nc);
 * 
 * nr = number of rows;
 * 
 * nc = number of columns;
 * 
 * theRow = a vector of length nr*nc that contains the constraint upon exit
 * 
 */

#define kEdgeWeight 1.25
#define kCornerWeight 1.67

void Get2Dsmoother(float smoother, float vwt, float hwt, int whichParameter, int nr, int nc, float *theRow)
{
	float           wing, center;
	int             i, whichRow, whichCol, offset;

	/* zero the row */

	for (i = 1; i <= nr * nc; i++)
		theRow[i] = 0.0;

	center = smoother * 2 * (vwt + hwt);
	wing = (-1) * smoother;

	/*
	 * Find the position of the parameter in the image The constraint is
	 * different for parameters located at a corner, edge, or within the
	 * interior of the cell
	 * 
	 */

	/* convert the cell "index" to an (r,c) pair */

	whichRow = (int) ((whichParameter - 1) / nc) + 1;
	whichCol = (whichParameter % nc);
	if (whichCol == 0)
		whichCol = nc;

	/*
	 * now find the appropriate constraint "offset" points to the center
	 * of the differencing star
	 */

	offset = whichParameter;

	/* a parameter along the top edge of the image */

	if (whichRow == 1) {

		/* the upper left corner */

		if (whichCol == 1) {

			theRow[offset] = center * kCornerWeight;
			theRow[offset + 1] = 2 * hwt * wing * kCornerWeight;
			theRow[offset + nc] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* the upper right corner */

		if (whichCol == nc) {

			theRow[offset] = center * kCornerWeight;
			theRow[offset - 1] = 2 * hwt * wing * kCornerWeight;
			theRow[offset + nc] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* parameters along the top, but not a corner */

		theRow[offset] = center * kEdgeWeight;
		theRow[offset - 1] = hwt * wing * kEdgeWeight;
		theRow[offset + 1] = hwt * wing * kEdgeWeight;
		theRow[offset + nc] = 2 * vwt * wing * kEdgeWeight;
		return;
	}
	/* parameter along the bottom edge of the model */

	if (whichRow == nr) {

		/* the lower left corner */

		if (whichCol == 1) {

			theRow[offset] = center * kCornerWeight;
			theRow[offset + 1] = 2 * hwt * wing * kCornerWeight;
			theRow[offset - nc] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* the lower right corner */

		if (whichCol == nc) {

			theRow[offset] = center * kCornerWeight;
			theRow[offset - 1] = 2 * hwt * wing * kCornerWeight;
			theRow[offset - nc] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* parameters along the bottom, but not a corner */

		theRow[offset] = center * kEdgeWeight;
		theRow[offset - 1] = hwt * wing * kEdgeWeight;
		theRow[offset + 1] = hwt * wing * kEdgeWeight;
		theRow[offset - nc] = 2 * vwt * wing * kEdgeWeight;
		return;
	}
	/* parameter along the left edge of the model (Corners already done) */

	if (whichCol == 1) {

		theRow[offset] = center * kEdgeWeight;
		theRow[offset + 1] = 2 * hwt * wing * kEdgeWeight;
		theRow[offset - nc] = vwt * wing * kEdgeWeight;
		theRow[offset + nc] = vwt * wing * kEdgeWeight;
		return;
	}
	/* parameter along the right edge of the model (Corners already done) */

	if (whichCol == 1) {

		theRow[offset] = center * kEdgeWeight;
		theRow[offset - 1] = 2 * hwt * wing * kEdgeWeight;
		theRow[offset - nc] = vwt * wing * kEdgeWeight;
		theRow[offset + nc] = vwt * wing * kEdgeWeight;
		return;
	}
	/* a parameter in the interior of the model  */

	theRow[offset] = center;
	theRow[offset - 1] = hwt * wing;
	theRow[offset + 1] = hwt * wing;
	theRow[offset - nc] = vwt * wing;
	theRow[offset + nc] = vwt * wing;

	return;

}


/*
 * Function to return a row of a "Laplacian" Smoothness constraint sub-matrix
 * for a 2D image
 * 
 * ASSUMES THAT THE IMAGE INDEXING BEGINS WITH ONE, SO THE FIRST ROW CONTAINS
 * ELEMENTS 1...NCOLS
 * 
 * This version returns only the nozero values of the row and their indices
 * 
 * the differencing star has:
 * 
 * 2(vwt+hwt)*smoother in the center
 * 
 * -vwt*smoother above/below
 * 
 * -hwt*smoother left/right
 * 
 * 
 * smoother = the smoothness weight value; vwt = the relative vertical
 * difference weight; hwt = the relative horizontal weight; whichParameter =
 * the index of the "parameter" to be constrained; nr = number of rows; nc =
 * number of columns; theRow = a vector of length nr*nc that contains the
 * constraint upon exit
 * 
 */

#define kEdgeWeight 1.25
#define kCornerWeight 1.67

void Get2DsmootherValues(float smoother, float vwt, float hwt, int whichParameter, int nr, int nc, float *theRow, int *indices, int *nnot_zero)
{
	float           wing, center;
	int             i, whichRow, whichCol, offset;

	center = smoother * 2 * (vwt + hwt);
	wing = (-1) * smoother;

	/*
	 * Find the position of the parameter in the image The constraint is
	 * different for parameters located at a corner, edge, or within the
	 * interior of the cell
	 * 
	 */

	/* convert the cell "index" to an (r,c) pair */

	whichRow = (int) ((whichParameter - 1) / nc) + 1;
	whichCol = (whichParameter % nc);
	if (whichCol == 0)
		whichCol = nc;

	/*
	 * now find the appropriate constraint "offset" points to the center
	 * of the differencing star
	 */

	offset = whichParameter;

	/* a parameter along the top edge of the image */

	if (whichRow == 1) {

		/* the upper left corner */

		if (whichCol == 1) {

			*nnot_zero = 3;
			indices[1] = offset;
			theRow[1] = center * kCornerWeight;
			indices[2] = offset + 1;
			theRow[2] = 2 * hwt * wing * kCornerWeight;
			indices[3] = offset + nc;
			theRow[3] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* the upper right corner */

		if (whichCol == nc) {
			*nnot_zero = 3;
			indices[1] = offset;
			theRow[1] = center * kCornerWeight;
			indices[2] = offset - 1;
			theRow[2] = 2 * hwt * wing * kCornerWeight;
			indices[3] = offset + nc;
			theRow[3] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* parameters along the top, but not a corner */

		*nnot_zero = 4;
		indices[1] = offset;
		theRow[1] = center * kEdgeWeight;
		indices[2] = offset - 1;
		theRow[2] = hwt * wing * kEdgeWeight;
		indices[3] = offset + 1;
		theRow[3] = hwt * wing * kEdgeWeight;
		indices[4] = offset + nc;
		theRow[4] = 2 * vwt * wing * kEdgeWeight;
		return;
	}
	/* parameter along the bottom edge of the model */

	if (whichRow == nr) {

		/* the lower left corner */

		if (whichCol == 1) {

			*nnot_zero = 3;
			indices[1] = offset;
			theRow[1] = center * kCornerWeight;
			indices[2] = offset + 1;
			theRow[2] = 2 * hwt * wing * kCornerWeight;
			indices[3] = offset - nc;
			theRow[3] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* the lower right corner */

		if (whichCol == nc) {

			*nnot_zero = 3;
			indices[1] = offset;
			theRow[1] = center * kCornerWeight;
			indices[2] = offset - 1;
			theRow[2] = 2 * hwt * wing * kCornerWeight;
			indices[3] = offset - nc;
			theRow[3] = 2 * vwt * wing * kCornerWeight;
			return;
		}
		/* parameters along the bottom, but not a corner */

		*nnot_zero = 4;
		indices[1] = offset;
		theRow[1] = center * kEdgeWeight;
		indices[2] = offset - 1;
		theRow[2] = hwt * wing * kEdgeWeight;
		indices[3] = offset + 1;
		theRow[3] = hwt * wing * kEdgeWeight;
		indices[4] = offset - nc;
		theRow[4] = 2 * vwt * wing * kEdgeWeight;
		return;
	}
	/* parameter along the left edge of the model (Corners already done) */

	if (whichCol == 1) {

		*nnot_zero = 4;
		indices[1] = offset;
		theRow[1] = center * kEdgeWeight;
		indices[2] = offset + 1;
		theRow[2] = 2 * hwt * wing * kEdgeWeight;
		indices[3] = offset - nc;
		theRow[3] = vwt * wing * kEdgeWeight;
		indices[4] = offset + nc;
		theRow[4] = vwt * wing * kEdgeWeight;
		return;
	}
	/* parameter along the right edge of the model (Corners already done) */

	if (whichCol == 1) {

		*nnot_zero = 4;
		indices[1] = offset;
		theRow[1] = center * kEdgeWeight;
		indices[2] = offset - 1;
		theRow[2] = 2 * hwt * wing * kEdgeWeight;
		indices[3] = offset - nc;
		theRow[3] = vwt * wing * kEdgeWeight;
		indices[4] = offset + nc;
		theRow[4] = vwt * wing * kEdgeWeight;
		return;
	}
	/* a parameter in the interior of the model  */

	*nnot_zero = 5;
	indices[1] = offset;
	theRow[1] = center;
	indices[2] = offset - 1;
	theRow[2] = hwt * wing;
	indices[3] = offset + 1;
	theRow[3] = hwt * wing;
	indices[4] = offset - nc;
	theRow[4] = vwt * wing;
	indices[5] = offset + nc;
	theRow[5] = vwt * wing;

	return;

}
#undef kEdgeWeight
#undef kCornerWeight

/*
 * Function to append a smoothness submatrix to a file without bothering with
 * the zero values.  Only the nonzero values are returned from
 * Get2DsmootherValues and only those values are written to the disk.
 * 
 */

/*
 * Function to return a row of a Smoothness constraint sub-matrix for a 2D
 * image - smoothness is distance dependent
 * 
 * ASSUMES THAT THE IMAGE INDEXING BEGINS WITH ONE, SO THE FIRST ROW CONTAINS
 * ELEMENTS 1...NCOLS
 * 
 * This version returns only the nonzero values of the row and their indices
 * 
 */


void GetDsmootherValues(float smoother, float maxdist, int whichParameter, int nr, int nc, float *theRow, int *indices, int *nnot_zero)
{

	float           dist;
	int             i, itmp, whichRow, whichCol, r, c, idst_squared;
	int		idist, rmin, rmax, cmin, cmax;

	if (maxdist < 1) {
		fprintf(stderr, "GetDSmoother: Something wrong with smoothness constraints.\n");
		fprintf(stderr, "GetDSmoother: Distance for smoothing less than one cell.\n");
		return;
	}
	/* work with an integer for speed */
	idist = (int) (maxdist + 0.5);
	idst_squared = (int) ((maxdist * maxdist) + 0.5);

	/* convert the cell "index" to an (r,c) pair */

	whichRow = (int) ((whichParameter - 1) / nc) + 1;
	whichCol = (whichParameter % nc);
	if (whichCol == 0)
		whichCol = nc;


	/* now find the cells within maxdist of the cell of interest */
	
	*nnot_zero = 0;

	rmin = whichRow - idist;
	if (rmin < 1)
		rmin = 1;
	rmax = whichRow + idist;
	if (rmax > nr)
		rmax = nc;

	cmin = whichCol - idist;
	if (cmin < 1)
		cmin = 1;
	cmax = whichCol + idist;
	if (rmax > nc)
		cmax = nc;

	for (r = rmin; r <= rmax; r++) {
		for (c = cmin; c <= cmax; c++) {

			/*
			 * compute the distance between this cell and the
			 * cell of interest
			 */
			itmp = (whichRow - r) * (whichRow - r) + (whichCol - c) * (whichCol - c);

			/* if acceptable, add the smoothness constraint */
			if (itmp <= idst_squared) {
				*nnot_zero += 1;
				theRow[*nnot_zero] = -1;
				indices[*nnot_zero] = ((r-1)*nc) + c;
				if (itmp == 0)
					theRow[*nnot_zero] = 1;
			}
		}
	}

	/* normalize so the the sum of the weights equals zero */

	for (i = 1; i <= *nnot_zero; i++) {
		if (i != whichParameter) {
			theRow[i] /= (float) (*nnot_zero - 1);
		}
	}


	return;

}

/*
 * Function to append a smoothness submatrix to a file without bothering with
 * the zero values.  Only the nonzero values are returned from
 * Get2DsmootherValues and only those values are written to the disk.
 * 
 */



#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>


void Append2Dsmoother(FILE *intfile, FILE *flofile, float vwt, float hwt, int nr, int nc, float smoother)
{
	int            *theIndices, nnot_zero, i, j;
	float          *theFloats;


	theFloats = (float *) malloc(5 * sizeof(float));
	theIndices = (int *) malloc(5 * sizeof(int));
	theFloats -= 1;
	theIndices -= 1;


	for (i = 1; i <= nr * nc; i++) {

		Get2DsmootherValues(smoother, vwt, hwt, i, nr, nc, theFloats, theIndices, &nnot_zero);

		/* adjust so that indexing begins with zero */

		for (j = 1; j <= nnot_zero; j++)
			theIndices[j] -= 1;

		/* write out the nonzero values */

		PutRowValues(intfile, flofile, &theFloats[1], &theIndices[1], &nnot_zero);
	}


	theFloats += 1;
	theIndices += 1;
	free((char *) theFloats);
	free((char *) theIndices);

	return;
}


#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>


void MultiAppend2Dsmoother(FILE *intfile, FILE *flofile, float vwt, float hwt, int nr, int nc, float smoother,int nImages)
{
	int            *theIndices, nnot_zero, i, j, k, nCells;
	float          *theFloats;


	nCells = nr*nc;
	theFloats = (float *) malloc(5 * sizeof(float));
	theIndices = (int *) malloc(5 * sizeof(int));
	theFloats -= 1;
	theIndices -= 1;

	for (k = 0; k < nImages; k++)
	  {
	for (i = 1; i <= nr * nc; i++) {

		Get2DsmootherValues(smoother, vwt, hwt, i, nr, nc, theFloats, theIndices, &nnot_zero);

		/* adjust so that indexing begins with zero */

		for (j = 1; j <= nnot_zero; j++)
			theIndices[j] -= 1;

		/* adjust the smoothing values so that they correspond to the correct image in the model */
		for(j=1;j<=nnot_zero;j++)
		  theIndices[j] += k * nCells;

		/* write out the nonzero values */

		PutRowValues(intfile, flofile, &theFloats[1], &theIndices[1], &nnot_zero);
	}
	  }

	theFloats += 1;
	theIndices += 1;
	free((char *) theFloats);
	free((char *) theIndices);

	return;
}
