/*****************************************************************

 October 1996
 added weighting  - use index of 1 for weights 
 do not weight the observations outside this routine, weighting
is applied  accomplished here

Routine to solve ax = b using
the algorithm lsqr of Paige and
Saunders as presented in Nolet.

Damping is applied using the methods in the
original fortran routines of paige and saunders.

The matrix a is stored in the file pointed at by fd.

fd is the file descriptor
nr = number of rows
nc = number of columns
x = the vector of unknowns (dimension = nc)
b = the vector of observations (dmimension = nr)
max_iterations - number of iterations in LSQR algebra
damper = damping factor for the eqns
max_cond_num = maximum condition number allowed (stop if exceeded).
residual = residual estimated from solution

****************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <sys/file.h>
#include <stdlib.h>
#include <math.h>
#include "invsubs.h"

void WtDiskLSQR(char *fn, int nr, int nc, float *x, float *b, int max_iterations, float damper, float max_cond_num, float *residual,float *wt)
{

	int             i, iteration;
	float           alpha, beta, length;
	float           rhobar, phibar, rho, phi;
	float           c, s, t1, t2, teta, dummy;
	float          *v, *w;
	float           res1, res2, rhbar1, rhbar2, cs1, sn1, psi;
	float           t3, tau, ddnorm, bbnorm, acond, anorm, xnorm;
	float           cs, sn, theta;
	float		oldres;
	
	oldres = 0;

	fprintf(stdout, "\nFrom PSLSQR:\n");
	fprintf(stdout, "nr = %d, nc = %d, damper = %f\n", nr, nc, damper);
	fprintf(stdout, "maximum condition number allowed is = %f\n", max_cond_num);

	/*
	 * allocate the work arrays and subtract 1 from the pointer so we can
	 * access the first element as v[1] not v[0]
	 */
	v = (float *) malloc((unsigned) (nc * sizeof(float)));
	w = (float *) malloc((unsigned) (nc * sizeof(float)));
	v -= 1;
	w -= 1;

	res1 = 0;
	res2 = 0;
	bbnorm = 0;
	ddnorm = 0;
	acond = 0;

	for (i = 1; i <= nc; i++) {
		x[i] = 0.0;
		v[i] = 0.0;
	}

        /* weight the observations */
	for(i=1;i<=nr;i++)
		b[i] *= wt[i];

	Normalize(b, nr, &length);
	beta = length;

	WtDiskATUPV(fn, nr, nc, b, v, wt);
	Normalize(v, nc, &length);
	alpha = length;

	rhobar = alpha;
	phibar = beta;

	for (i = 1; i <= nc; i++)
		w[i] = v[i];

	for (iteration = 0; iteration < max_iterations; iteration++) {
		dummy = -alpha;
		for (i = 1; i <= nr; i++)
			b[i] *= dummy;

		WtDiskAVPU(fn, nr, nc, b, v, wt);

		Normalize(b, nr, &length);
		beta = length;
		bbnorm = bbnorm + alpha * alpha + beta * beta + damper * damper;

		dummy = -beta;
		for (i = 1; i <= nc; i++)
			v[i] *= dummy;

		WtDiskATUPV(fn, nr, nc, b, v, wt);

		Normalize(v, nc, &length);
		alpha = length;

		rhbar2 = rhobar * rhobar + damper * damper;
		rhbar1 = sqrt(rhbar2);
		cs1 = rhobar / rhbar1;
		sn1 = damper / rhbar1;
		psi = sn1 * phibar;
		phibar = cs1 * phibar;

		rho = sqrt((double) (rhbar2 + beta * beta));
		cs = rhbar1 / rho;
		sn = beta / rho;
		theta = sn * alpha;
		rhobar = -cs * alpha;
		phi = cs * phibar;
		phibar = sn * phibar;
		tau = sn * phi;

		t1 = phi / rho;
		t2 = -theta / rho;
		t3 = 1 / rho;

		xnorm = 0;
		for (i = 1; i <= nc; i++) {
			x[i] += (t1 * w[i]);
			xnorm += x[i] * x[i];
			w[i] *= t2;
			w[i] += v[i];
			dummy = w[i];
			dummy *= dummy * t3 * t3;
			ddnorm += dummy;
		}
		xnorm = sqrt(xnorm);

		anorm = sqrt(bbnorm);
		acond = anorm * sqrt(ddnorm);
		res1 = phibar * phibar;
		res2 += psi * psi;
		*residual = sqrt(res1 + res2);
		fprintf(stdout, "Iteration %2d:  cond # = %f, xnorm = %f, Res = %f\n", iteration + 1, acond, xnorm, *residual);

		if (acond >= max_cond_num ) {
			iteration = max_iterations;
			fprintf(stdout, "Condition Number too large, halting iteration.\n");
		}

		if (fabs(oldres - *residual) < 1e-06 ) {
			iteration = max_iterations;
			fprintf(stdout, "Residual has converged - halting iteration.\n");
		}
				
		oldres = *residual;
		
	}

	w += 1;
	v += 1;
	free((char *) w);
	free((char *) v);

}
/* end pslsqr */


/******** compute u = u + a*v  ************/

void WtDiskAVPU(char *fn, int nr, int nc, float *u, float *v, float *wt)
{
	int             r, c, nBytes_row, *indices, nnot_zero, r_error;
	float          *a_row, *atimesv, sum;
	FILE           *intfile, *flofile;

	/* open the matrix files */

	intfile = fopen("L_indices", "r");
	flofile = fopen("L_floats", "r");

	/* allocate the memory for two rows */

	nBytes_row = nc * sizeof(float);

	a_row = (float *) malloc((unsigned) nBytes_row);
	a_row -= 1;
	indices = (int *) malloc((unsigned) nc * sizeof(int));
	indices -= 1;
	atimesv = (float *) malloc((unsigned) nr * sizeof(float));
	atimesv -= 1;

	/*
	 * read in a row at a time and compute A * v
	 */

	for (r = 1; r <= nr; r++) {

		GetRowValues(intfile, flofile, &a_row[1], &indices[1], &nnot_zero, &r_error);

		sum = 0;
		if (nnot_zero != 0) {
			for (c = 1; c <= nnot_zero; c++)
				sum += (a_row[c] * v[indices[c] + 1]);
		}
		atimesv[r] = (float) sum * wt[r];
	}

	fclose(intfile);
	fclose(flofile);

	/*
	 * compute  u + A * v
	 * 
	 */

	for (r = 1; r <= nr; r++)
		u[r] = u[r] + atimesv[r];

	/* reset the pointer  and free the memory */

	a_row += 1;
	atimesv += 1;
	indices += 1;

	free((char *) atimesv);
	free((char *) a_row);
	free((char *) indices);

}

/****** compute v = v + a transp * u *******/

void WtDiskATUPV(char *fn, int nr, int nc, float *u, float *v, float *wt)
{
	int             r, c, nBytes_row, *indices, nnot_zero,r_error;
	float          *a_row, *atu, dumfl;


	FILE           *intfile, *flofile;

	intfile = fopen("L_indices", "r");
	flofile = fopen("L_floats", "r");


	nBytes_row = nc * sizeof(float);
	a_row = (float *) malloc((unsigned) nBytes_row);
	a_row -= 1;
	indices = (int *) malloc((unsigned) nc * sizeof(int));
	indices -= 1;
	atu = (float *) malloc((unsigned) nBytes_row);
	atu -= 1;

	for (c = 1; c <= nc; c++)
		atu[c] = 0;

	/*
	 * read in a row at a time and compute A**T*u
	 * 
	 */

	for (r = 1; r <= nr; r++) {

		GetRowValues(intfile, flofile, &a_row[1], &indices[1], &nnot_zero,&r_error);

		dumfl = u[r] * wt[r]; /* this wt is from each row */

		if (nnot_zero != 0) {
			for (c = 1; c <= nnot_zero; c++)
				atu[indices[c] + 1] += (a_row[c] * dumfl);
		}
	}

	fclose(intfile);
	fclose(flofile);

	/* now add A**T * u to the vector v */

	for (c = 1; c <= nc; c++)
		v[c] = v[c] + atu[c];

	/* reset the pointers and free the memory */

	a_row += 1;
	atu += 1;
	indices += 1;

	free((char *) a_row);
	free((char *) atu);
	free((char *) indices);

}
