/****************************************************************************************/
/*
 * Program: multi_lsqr Version: 1.0
 * 
 * program to invert group velocities using great-circle paths This code uses a
 * conjugate gradient algorithm (LSQR) to perform an approximate inversion of
 * the matrix equations.
 * 
 * Chuck Ammon St. Louis University, October 1996 (nmlsqr)
 *    reworked under Sun's workshop November 2002 (nmlsqr)
 *    converted to multiple-periods    March 2003 (mnmlsqr)
 *    moved to the Mac                 June  2003 (mult_lsqr)
 * 
 * Some Details:
 * 
 * The inversion grid is set up to be uniform length cells (measured in km)
 * assuming a sinusoidal projection). The output is ascii percent velocity
 * perturbations, the fast and the slow values are written into separate
 * files. The procedure for the inversion is iterative, but the ray paths are
 * fixed during the inversion.
 * 
 * The input are in ascii format and consist of the event and station locations
 * and the phase velocity.  The initial model is a constant velocity model,
 * which you must also supply.
 * 
 * Laplacian smoothness constraints are included and so is an aprior model
 * constraint (minimize the difference between the solution and a specified
 * model), which right now is limited to using the constant velocity model.
 * 
 * 
 */
/****************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include "invsubs.h"
#include "libget.h"

#define deg_to_rad 0.0174532
#define kSmall 0.0001

/****************************************************************************/
/* define a structure to hold the observations */
/****************************************************************************/


/* prototypes */
void getStats(float *x, int n, float *mean, float *variance);
void GetModelRough(int m,float *s,int nx,int ny,float *meanvel, float *roughvel);
struct OBSERVATION nReadAnObservation(FILE *fd0);

/****************************************************************************/

int main(ac, av)
	int             ac;
	char          **av;
{
	float           az,weightCutoff=4.0;
	float          *l, *l0;
	float          *v, *s, *ds, *dc_velocity;
	float           time_tol = 0.0001;
	float           small = .00005;
	float           damper = 0.1;
	float           minres = 2.0;
	float           sum = 0, res, predicted;
	float          *plats, *plons, *pYs, *pXs, smple_dist, max_cond_num = 100;
	float           recs_constraint = 1;
	float           Wx, Wy;
	float           minlat, minlon, observed, rms;
	float           *initialPeriodMisfits, *finalPeriodMisfits, *sumWtsPerPeriod,temp;
	float           center_lon, eastern, western, northern, southern, min_x;
	float           lsqr_res, corr_wt = 1;
	float           maxdist = 0, mindist = 0;
	float           min_res_bin, res_bin_width;
	float           sum_of_wts = 0, epsilon = 1.0;
	float           centerlat, so_far, fraction_done;
	float           arc_distance, distance, *observations, smoother = 1, apmodelwt = 0;
	float           vel_res, vel_sum, vel_rms,initialRMS,initialVelocityRMS;
	float          *theResiduals,theInitialMean,theInitialVariance,theFinalMean,theFinalVariance;
	float           *ltt, *ltl, *modelLength, *modelRoughness, rough;
	float           minPeriod, dPeriod, distcheck, suggestedWeight, wres;


	char            obsfile[256], resfile[256], pathMatrix[256];
	FILE           *rcfd;
	FILE           *intfile, *flofile;
	FILE           *fd0, *fd1, *fd2, *logFile;

	int             nCells, nCellsPerImage, nImages, verbose = 0, write_times = 0;
	int             nx, ny, i, j, k, n, fd;
	int             numRows, nobs, iperiod, ioffset, *nObsPerPeriod;
	int             lsqr_iter = 20, iobs;
	int             iteration, max_iteration = 1;
	int             nsegs, mx_segs = 20000;
	int            *indices, nnot_zero;
	int             computePaths = 1, r_error;
	int             nwritten = 0;
	int            *resbins, nres_bins, ires_bin;
	int             iwt = 0;
	int             keep = 1, ekeep = 1, mask = 0, minobs = 4;

float dumpLat, dumpLon, dumpRadDeg;

	float *Period;
	float meanvel, roughvel;
	struct OBSERVATION anObservation, nReadAnObservation();

	logFile = fopen("SRFLSQR.log", "w");
	setlinebuf(logFile);
	strcpy(pathMatrix, "DUMMY");

	/**********************************************************************************/
	/* this value is used to sample along the path - it is in km */
	/* the max path length is smpl_dist * mx_segs, which is 20,000km right now */
	/**********************************************************************************/
	smple_dist = 1.0;
	mx_segs = 20000;

	/*****************************************************/
	/* set up a vector to hold the binned residual stats */
	/*****************************************************/
	res_bin_width = 2;
	nres_bins = 500;
	min_res_bin = -501;
	resbins = (int *) calloc(nres_bins, sizeof(int));

	/*****************************************************/
	/* User input: */
	/**************************************************** */
	setpar(ac, av);
	getpar("max_iteration", "d", &max_iteration);
	/*mstpar("dc_velocity", "f", &dc_velocity);*/
	mstpar("xwidth", "f", &Wx);
	mstpar("ywidth", "f", &Wy);
	mstpar("minlat", "f", &minlat);
	mstpar("minlon", "f", &minlon);
	mstpar("nx", "d", &nx);
	mstpar("ny", "d", &ny);
	mstpar("nobs", "d", &nobs);
	mstpar("obsfile", "s", obsfile);
	getpar("verbose", "d", &verbose);
	getpar("predict", "d", &write_times);
	getpar("damper", "f", &damper);
	getpar("computepaths", "d", &computePaths);
	computePaths = 1; /* the optimal dc_velocity is calculated while computing paths */
	getpar("epsilon", "f", &epsilon);
	getpar("iwtevents", "d", &iwt);
	getpar("max_cond_num", "f", &max_cond_num);
	getpar("lsqr_iter", "d", &lsqr_iter);
	getpar("smoother", "f", &smoother);
	getpar("apmodelwt", "f", &apmodelwt);
	getpar("weightCutoff", "f", &weightCutoff);
	mstpar("nImages", "d", &nImages); /* i.e. nPeriods (or later, R & G) */

	/* RBH add 2015 05 25 */
	getpar("dumpLat", "f", &dumpLat );
	getpar("dumpLon", "f", &dumpLon );
	getpar("dumpRadDeg", "f", &dumpRadDeg );
	/* end RBH */


fprintf(stderr,"epsilon         %f\n",epsilon);
fprintf(stderr,"max_iteration %d\n",max_iteration);
fprintf(stderr,"xwidth          %f\n",Wx);
fprintf(stderr,"ywidth          %f\n",Wy);
fprintf(stderr,"minlat          %f\n",minlat);
fprintf(stderr,"minlon          %f\n",minlon);
fprintf(stderr,"nx              %d\n",nx);
fprintf(stderr,"ny              %d\n",ny);
fprintf(stderr,"nobs            %d\n",nobs);
fprintf(stderr,"obsfile         %s\n",obsfile);
fprintf(stderr,"verbose         %d\n",verbose);
fprintf(stderr,"predict         %d\n",write_times);
fprintf(stderr,"damper          %f\n",damper);
fprintf(stderr,"iwtevents       %d\n",iwt);
fprintf(stderr,"computepaths    %d\n",computePaths);
fprintf(stderr,"max_cond_num    %f\n",max_cond_num);
fprintf(stderr,"lsqr_iter       %d\n",lsqr_iter);
fprintf(stderr,"smoother        %f\n",smoother);
fprintf(stderr,"apmodelwt       %f\n",apmodelwt);
fprintf(stderr,"weightCutoff    %f\n",weightCutoff);
fprintf(stderr,"nImages         %d\n",nImages);
	/*********************************************************************************/
	/* Echo the User input */
	/*********************************************************************************/
	fprintf(logFile, "xwidth=%f ywidth=%f\n", Wx, Wy);
	fprintf(logFile, "minlat=%f minlon=%f\n", minlat, minlon);
	fprintf(logFile, "nx=%d ny=%d nobs=%d\n", nx, ny, nobs);
	fprintf(logFile, "nobs=%d\n", nobs);
	fprintf(logFile, "obsfile=%s\n", obsfile);
	fprintf(logFile, "damper=%f\n", damper);
	if (iwt == 1)
		fprintf(stdout, "Weighting the events using event bins.\n");
        /*********************************************************************************/
	/* Allocate memory for the other vectors and arrays  */
	/*********************************************************************************/
	nCellsPerImage = nx * ny;
	nCells = nCellsPerImage * nImages;

	observations = (float *) calloc(nobs + 2 * nCells, sizeof(float));
	theResiduals = (float *) calloc(nobs, sizeof(float));

	l0      = (float *) calloc(nCellsPerImage, sizeof(float));
	l       = (float *) calloc(nCells, sizeof(float));
	indices = (int *)   calloc(nCells, sizeof(int));
	ds      = (float *) calloc(nCells, sizeof(float));
	s       = (float *) calloc(nCells + 1, sizeof(float));
	v       = (float *) calloc(nCells, sizeof(float));
	plats   = (float *) calloc(mx_segs, sizeof(float));
	plons   = (float *) calloc(mx_segs, sizeof(float));
	pXs     = (float *) calloc(mx_segs, sizeof(float));
	pYs     = (float *) calloc(mx_segs, sizeof(float));

	ltt =         (float *) calloc(nImages, sizeof(float));
	ltl =         (float *) calloc(nImages, sizeof(float));
	dc_velocity = (float *) calloc(nImages, sizeof(float));
	modelLength = (float *) calloc(nImages, sizeof(float));
	modelRoughness = (float *) calloc(nImages, sizeof(float));
	initialPeriodMisfits = (float *) calloc(nImages, sizeof(float));
	finalPeriodMisfits = (float *) calloc(nImages, sizeof(float));
	sumWtsPerPeriod = (float *) calloc(nImages, sizeof(float));
	nObsPerPeriod = (int *) calloc(nImages, sizeof(float));
	Period = (float *) calloc(nImages, sizeof(float));

	/****************************************************************************/
	/* set up some constants for the grid. */
	/****************************************************************************/

	northern = minlat + (ny * Wy) / 111.19;
	southern = minlat;
	western = minlon;
	eastern = minlon + (nx * Wx) / 111.19;
	center_lon = 0.5 * (eastern + western);
	min_x = (center_lon - western) * 111.19;
	centerlat = 0.5 * (southern + northern);
	fprintf(logFile, "Grid:\nsouthern,northern,western,eastern\n");
	fprintf(logFile, "%f,%f,%f,%f\n", southern, northern, western, eastern);
        WriteModelGrid(minlat,minlon,Wx,Wy,nx,ny,center_lon,min_x);

	iteration = 1;

	if (computePaths) {
		/****************************************************************************/
		/* open the pathmatrix files for output of ALL Ray paths */
		/****************************************************************************/

		intfile = fopen("L_indices", "w");
		if (intfile == NULL) {
			fprintf(stderr, "Problem opening the path index file.\n");
			exit(-1);
		}
		flofile = fopen("L_floats", "w");
		if (flofile == NULL) {
			fprintf(stderr, "Problem opening the path float file.\n");
			exit(-1);
		}
		nwritten = 0;

		/****************************************************************************/
		/* open up the observation and predictions files */
		/****************************************************************************/

		fd0 = fopen(obsfile, "r");
		if (fd0 == NULL) {
			fprintf(stderr, "Problem opening file: %s\n", obsfile);
			exit(-1);
		}
		/****************************************************************************/
		/* loop over the observations */
		/****************************************************************************/

		if (verbose) {
			fprintf(stdout, "------------------------------\n");
			fprintf(stdout, "Computing the ray paths\n");
			fprintf(stdout, "------------------------------\n");
		}
		fraction_done = 0.05;
		fprintf(stderr, "xxxxxxxxxxxxxxxxxxxx\n");

		for (iobs = 0; iobs < nobs; iobs++) {

			anObservation = nReadAnObservation(fd0);
/*
			anObservation.iperiod = (int) ((anObservation.period - minPeriod) / dPeriod);
*/
			Period[anObservation.iperiod] = anObservation.period ;

			delta_az(&anObservation.evlat, &anObservation.evlon, &anObservation.stlat, &anObservation.stlon, &arc_distance, &az);
			arc_distance *= 111.19;
			nsegs = (arc_distance / smple_dist) + 1;

			/* used to compute the least-squares slowness */
			ltl[anObservation.iperiod] += arc_distance * arc_distance *anObservation.weight*anObservation.weight;
			ltt[anObservation.iperiod] += arc_distance * arc_distance *anObservation.weight*anObservation.weight / anObservation.phs_velocity;

			if (nsegs > mx_segs)
				nsegs = mx_segs;

			Gcarc(anObservation.evlat, anObservation.evlon, anObservation.stlat, anObservation.stlon, az, arc_distance, plats, plons, nsegs);

			for (i = 0; i < nsegs; i++) {
/*
DEBUG
*/
if ( dumpRadDeg > 0 ){
	if( fabs(plats[i] - (dumpLat) ) < 1.0 && fabs(plons[i] - (dumpLon)) < (dumpRadDeg) )
	{
	fprintf(stderr,"O: %d %d [%f,%f]  (%f,%f) (%f,%f) %f %f %f\n",i,iobs,plats[i],plons[i],anObservation.evlat, anObservation.evlon,anObservation.stlat, 
				anObservation.stlon,arc_distance,anObservation.period,anObservation.phs_velocity);
	} 
}

				LATLON_to_XY(plats[i], plons[i], &pXs[i], &pYs[i], minlat, center_lon, min_x);
/*
	printf("nsegs %d %d %f %f %f %f\n",nsegs,i,plats[i], plons[i],pXs[i],pYs[i]);
*/
			}

			GetGCARCLengths(pXs, pYs, plats, plons, nsegs, nx, ny, Wx, Wy, l0);

			ioffset = anObservation.iperiod * nCellsPerImage;

			for(i=0;i<nCells;i++)
			  l[i] = 0;

			for(i = 0; i< nCellsPerImage;i++)
			  l[i + ioffset] = (anObservation.weight * l0[i]);


			PutrowP(intfile, flofile, l, nCells, (float) 0.1);
			nwritten += 1;

			/* user feedback on how long to go */
			so_far = (float) iobs / (float) nobs;
			if (so_far > fraction_done) {
				fprintf(stderr, "x");
				fraction_done += 0.05;
			}
		}		/* end of loop over observations */
		fprintf(stdout, "o\n");
		fclose(intfile);
		fclose(flofile);
		fclose(fd0);
		fprintf(stdout, "Number of rows written: %d\n", nwritten);
	}

	/* finished with the path computations */


	/* compute the least squares slowness value and store its inverse in dc_velocity  */
	fd2 = fopen("mean_velocity.txt","w");
	fprintf(fd2,"____________________________\n");
	fprintf(fd2,"\nInitial DC Velocities\n");
	fprintf(fd2,"____________________________\n");
	for(j=0;j<nImages;j++)
	  {
	    dc_velocity[j] = ltl[j] / ltt[j];
	    fprintf(fd2,"%5.2f %10.4f\n",Period[j], dc_velocity[j]);
	    /****************************************************************************/
	    /* Initialize the velocity model */
	    /****************************************************************************/
	    for (i = 0; i < nCellsPerImage; i++) {
	           k = i + j * nCellsPerImage;
	        s[k] = 1 / dc_velocity[j];
	        v[k] = dc_velocity[j];
	       ds[k] = 0;
	    }
	  }
	fprintf(fd2,"____________________________\n");
	fclose(fd2);

	/********************************************************************************/
	/*
	 * NOW INVERT THE TRAVEL TIMES FOR LATERAL VELOCITY VARIATIONS
	 */
	/********************************************************************************/
	printf("Inverting for slowness perturbations.\n");

	intfile = fopen("L_indices", "r");
	if (intfile == NULL) {
		fprintf(stderr, "Problem opening the path index file.\n");
		fprintf(stderr, "Check the computepaths input parameter.\n");
		fprintf(stderr, "I'm quitting.\n");
		exit(-1);
	}
	flofile = fopen("L_floats", "r");
	if (flofile == NULL) {
		fprintf(stderr, "Problem opening the path float file.\n");
		exit(-1);
	}
	/********************************************************************************/
	/* open the observation file */
	/********************************************************************************/

	fd0 = fopen(obsfile, "r");
	if (fd0 == NULL) {
		fprintf(stderr, "Problem opening file: %s\n", obsfile);
		exit(-1);
	}
	numRows = 0;
	sum = 0;
	vel_sum = 0;

	fd2 = fopen("initialFits.txt","w");
	fprintf(fd2, "id evlat evlon stlat stlon period velocity observed predicted res %%res vel_res distance distCheck\n");
	/***************************************************************************************************************************/
	/* set up the vector with the times or observations storage */
	/***************************************************************************************************************************/
	if (verbose) {
		fprintf(stdout, "Converting group velocities to travel times.\n");
	}
	for (iobs = 0; iobs < nobs; iobs++) {

		anObservation = nReadAnObservation(fd0);
/*
		anObservation.iperiod = (int) ((anObservation.period - minPeriod) / dPeriod);
*/
		delta_az(&anObservation.evlat, &anObservation.evlon, &anObservation.stlat, &anObservation.stlon, &arc_distance, &az);
		arc_distance *= 111.19;
		observed = arc_distance / anObservation.phs_velocity;


		GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
		if (r_error == -1)
			fprintf(logFile, "Problem reading in row values: iobs=%d\n", iobs);

		predicted = 0;
		distance = 0;
		for (j = 0; j < nnot_zero; j++) {
			predicted += l[j] * s[indices[j]];
			ds[indices[j]] += l[j];	/* store the "coverage" in ds (still includes weighting factor)*/
			distance += l[j];
		}
		predicted /= anObservation.weight;
		distance  /= anObservation.weight;

		res = observed - predicted;
		distcheck = fabs(distance - arc_distance)/arc_distance;
		
		observations[numRows] = anObservation.weight * res;

		vel_res = arc_distance * (1 / observed - 1 / predicted);


		fprintf(fd2, "%s %10.2f %10.2f %10.2f %10.2f %8.2f %8.3f %10.3f %10.3f %10.3f %10.3f %8.4f %10.1f %10.4f\n",
			anObservation.event_id,
			anObservation.evlat, anObservation.evlon, anObservation.stlat,
			anObservation.stlon, anObservation.period, anObservation.phs_velocity, observed,  
			predicted, res, (res/observed)*100, vel_res, arc_distance, distcheck);


		ires_bin = (res - min_res_bin) / res_bin_width;

		if (ires_bin < 0 || ires_bin >= nres_bins)
			printf("Warning - residual out of bin range.\n");
		else
			resbins[ires_bin] += 1;

		sum += res * res;

		temp = res * anObservation.weight;
		temp *= temp;
		initialPeriodMisfits[anObservation.iperiod] += temp;
		sumWtsPerPeriod[anObservation.iperiod] += (anObservation.weight*anObservation.weight);
		nObsPerPeriod[anObservation.iperiod] += 1;

		vel_sum += vel_res * vel_res;
		theResiduals[numRows] = res;

		numRows += 1;

	}			/* end of loop over observations */

	fclose(fd0);
	fclose(fd2);
	fclose(intfile);
	fclose(flofile);

	getStats(theResiduals, numRows, &theInitialMean, &theInitialVariance);

	/********************************************************************************/
	/* compute the RMS errors in "Travel Time" and phase velocity */
	/********************************************************************************/
	rms = sqrt(sum / (float) numRows);
	vel_rms = sqrt(vel_sum / (float) numRows);
	fprintf(stdout, "\nInitial RMS = %f Velocity RMS = %f\n", rms, vel_rms);
	fprintf(logFile, "\nInitial RMS = %f Velocity RMS = %f\n", rms, vel_rms);
	initialRMS = rms; 
	initialVelocityRMS = vel_rms;

	/********************************************************************************/
	/* Error Binning */
	/********************************************************************************/

	sprintf(resfile, "%s.%2.2d", "resbins", iteration);
	rcfd = fopen(resfile, "w");
	for (i = 0; i < nres_bins; i++) {
		fprintf(rcfd, "%f %d\n", min_res_bin + (i + 0.5) * res_bin_width, resbins[i]);
	}
	fclose(rcfd);

	/* clear out the residual bins */
	for (i = 0; i < nres_bins; i++) {
		resbins[i] = 0;
	}

	/********************************************************************************/
	/* Append some a priori model constraints constraints */
	/********************************************************************************/
	if (apmodelwt > 0) {
		fprintf(stdout, "**************************\nAppending A Prior Model Equations\n");
		intfile = fopen("L_indices", "r+");
		if (intfile == NULL) {
			fprintf(stderr, "Problem opening the path index file.\n");
			exit(-1);
		}
		flofile = fopen("L_floats", "r+");
		if (flofile == NULL) {
			fprintf(stderr, "Problem opening the path float file.\n");
			exit(-1);
		}
		/* read past the ray paths */
		for (iobs = 0; iobs < nobs; iobs++) {
			GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
			if (r_error == -1)
				fprintf(logFile, "Problem reading in row values: iobs=%d\n", iobs);
		}

		j = 1;
		for (i = 0; i < nCells; i++) {
			PutRowValues(intfile, flofile, &apmodelwt, &i, &j);
			/* observations[numRows++] = apmodelwt / dc_velocity; */
			observations[numRows++] = apmodelwt;
		}
		

		fclose(intfile);
		fclose(flofile);
		/* numRows now equals nobs + nCells */
	}
	/********************************************************************************/
	/* Append some smoothing constraints */
	/********************************************************************************/

	if (smoother > 0) {
		fprintf(stdout, "Smoothing with a value of %f\n", smoother);
		fprintf(stdout, "Number of rows = %d\n", numRows);
		fprintf(stdout, "**************************\nAppending Smoothness Equations\n");
		fprintf(stderr, "Smoothing with a value of %f\n", smoother);
		fprintf(stderr, "Number of rows = %d\n", numRows);
		fprintf(stderr, "**************************\nAppending Smoothness Equations\n");
		intfile = fopen("L_indices", "r+");
		if (intfile == NULL) {
			fprintf(stderr, "Problem opening the path index file.\n");
			exit(-1);
		}
		flofile = fopen("L_floats", "r+");
		if (flofile == NULL) {
			fprintf(stderr, "Problem opening the path float file.\n");
			exit(-1);
		}
		/* read past the ray paths and the other constraints */
		for (iobs = 0; iobs < nobs; iobs++) {
			GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
			if (r_error == -1)
				fprintf(logFile, "Problem reading in row values: iobs=%d\n", iobs);
		}
		if (apmodelwt > 0) {
			for (iobs = 0; iobs < nCells; iobs++) {
				GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
				if (r_error == -1)
					fprintf(logFile, "Problem reading in row values: irow=%d\n", iobs+nobs);
			}
		}
		
		
		MultiAppend2Dsmoother(intfile, flofile, (float) 1, (float) 1, ny, nx, smoother, nImages);
		numRows += nCells;

		fclose(intfile);
		fclose(flofile);
	}
	/*************************************************************************************************/
	/* Invert the matrix equations for the new model, s[] */
	/*************************************************************************************************/

	for (i = 0; i < nCells + 1; i++) {
		s[i] = 0;
	}

	if (verbose)
		fprintf(stdout, "\nBeginning the LSQR computation.\n");

	observations -= 1;
	s -= 1;
	DiskLSQR(pathMatrix, numRows, nCells, s, observations, lsqr_iter, damper, max_cond_num, &lsqr_res);
	s += 1;
	observations += 1;

	/*************************************************************************************************/
	/* compute some statistics on the slowness perturbations, s[] */
	/*************************************************************************************************/

	if(smoother > 0.0)
	  {
	    for(k=0;k<nImages;k++)
	      {

		modelLength[k] = 0.0;
		for(i=0;i<nCellsPerImage;i++)
		  {
		    j = i + k * nCellsPerImage;
		    modelLength[k] += sqrt(s[j]*s[j]);
		  }
	    
		modelLength[k] /= (float) nCellsPerImage;
	    
		/* compute the model roughness */
		intfile = fopen("L_indices", "r+");
		if (intfile == NULL) {
		  fprintf(stderr, "Problem opening the path index file.\n");
		  exit(-1);
		}
		flofile = fopen("L_floats", "r+");
		if (flofile == NULL) {
		  fprintf(stderr, "Problem opening the path float file.\n");
		  exit(-1);
		}
		/* read past the ray paths and the other constraints */
		for (iobs = 0; iobs < nobs; iobs++) {
		  GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
		  if (r_error == -1)
		    fprintf(logFile, "Problem reading in row values: iobs=%d\n", iobs);
		}
		if (apmodelwt > 0) {
		  for (iobs = 0; iobs < nCells; iobs++) {
		    GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
		    if (r_error == -1)
		      fprintf(logFile, "Problem reading in row values: irow=%d\n", iobs+nobs);
		  }
		}
		/* now process the roughness */
		modelRoughness[k] = 0.0;
		for (iobs = 0; iobs < nCells; iobs++) {
		  GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
		  if (r_error == -1)
		    fprintf(logFile, "Problem reading in row values: irow=%d\n", iobs+nobs);
		  rough = 0.0;
		  for (j = 0; j < nnot_zero; j++) {
		    rough += (l[j] * s[indices[j]]);
		  }
		  modelRoughness[k] += sqrt(rough*rough);
		}
		modelRoughness[k] /= ((float) nCellsPerImage * smoother);
	    
		fclose(intfile);
		fclose(flofile);
	      }
	  }

	/*************************************************************************************************/
	/* convert ds to s */
	for(k=0;k<nImages;k++)
	{
	  ioffset = k * nCellsPerImage;
	  for(i=0;i<nCellsPerImage;i++)
	    s[i+ioffset] += (1/dc_velocity[k]);
	}


	if (smoother > 0)
		fprintf(logFile, "\nused %d out of %d rays, RMS = %f\n\n", numRows - nCells, nobs, rms);
	else
		fprintf(logFile, "\nused %d out of %d rays, RMS = %f\n\n", numRows, nobs, rms);


	/*************************************************************************************************/
	/* write the model into slow and fast ascii files */
	/*************************************************************************************************/

	WriteModels(iteration,ds,s,v,minlat,minlon,Wx,Wy,nx,ny,center_lon,min_x,nImages);

	/*************************************************************************************************/
	/* COMPUTE THE FINAL RESIDUALS */
	/*************************************************************************************************/

	if (verbose)
		fprintf(stdout, "Computing the final residuals.\n");

	intfile = fopen("L_indices", "r");
	if (intfile == NULL) {
		fprintf(stderr, "Problem opening the path index file.\n");
		exit(-1);
	}
	flofile = fopen("L_floats", "r");
	if (flofile == NULL) {
		fprintf(stderr, "Problem opening the path float file.\n");
		exit(-1);
	}
	/****************************************************************************/
	/* open up the observation and predictions files */
	/****************************************************************************/

	fd0 = fopen(obsfile, "r");
	if (fd0 == NULL) {
		fprintf(stderr, "Problem opening file: %s\n", obsfile);
		exit(-1);
	}
	fd1 = fopen("finalFits.txt", "w");
	if (fd1 == NULL) {
		fprintf(stderr, "Problem opening output file finalFits.txt.\n");
		exit(-1);
	}
	fprintf(fd1,"id evlat evlon stlat stlon period velocity observed predicted res %%res vel_res distance distCheck\n");
	
	fd2 = fopen("data.out", "w");
	if (fd2 == NULL) {
		fprintf(stderr, "Problem opening output file data.out.\n");
		exit(-1);
	}

	numRows = 0;
	sum = 0;
	sum_of_wts = 0;

	for(iobs=0;iobs<nobs;iobs++)
	  theResiduals[iobs] = 0.0;

	/*************************************************************************************************************************/
	/* compute the predictions - loop over ray paths */
	/*************************************************************************************************************************/

	for (iobs = 0; iobs < nobs; iobs++) {

		/* read in the data, one line at a time */

		anObservation = nReadAnObservation(fd0);
/*
		anObservation.iperiod = (int) ((anObservation.period - minPeriod) / dPeriod);
*/
		delta_az(&anObservation.evlat, &anObservation.evlon, &anObservation.stlat, &anObservation.stlon, &arc_distance, &az);
		arc_distance *= 111.19;
		observed = arc_distance / anObservation.phs_velocity;

		GetRowValues(intfile, flofile, &l[0], &indices[0], &nnot_zero, &r_error);
		if (r_error == -1)
			fprintf(logFile, "Problem reading in row values: iobs=%d\n", iobs);

		predicted = 0;
		distance  = 0;
		for (j = 0; j < nnot_zero; j++) {
			predicted += l[j] * s[indices[j]];
			distance  += l[j];
		}
		/* remove the effects of weighting in the "L" matrix */
		distance  /= anObservation.weight;
		predicted /= anObservation.weight;

		res = observed - predicted;
		wres = 100 * (res / observed);
		vel_res = anObservation.weight * arc_distance * (1 / observed - 1 / predicted);
		distcheck = fabs(distance - arc_distance) / arc_distance;
		
		suggestedWeight = 1.0;
		if(fabs(wres) > weightCutoff)
		  suggestedWeight = 1 / (4.*(fabs(wres)-weightCutoff) + 1);
/*
		  suggestedWeight = 1 / (fabs(wres)-weightCutoff + 1);
*/
                /* RBH 16 NOV 2016 the suggested weight is based on the residuals. Do not override the
 *                 a priori weight
		suggestedWeight *= anObservation.weight ;
		/* end RBH 16 NOV 2016 */

		fprintf(fd1, "%s %10.3f %10.3f %10.3f %10.3f %8.2f %10.3f %10.3f %10.3f %10.3f % 10.3f %10.3f %7.1f %10.4f\n",
			anObservation.event_id,
			anObservation.evlat, anObservation.evlon, anObservation.stlat,
			anObservation.stlon, anObservation.period,anObservation.phs_velocity, 
			observed, predicted, res, (res/observed)*100, vel_res,arc_distance,distcheck);

/* 		fprintf(fd1, "%-32s %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f\n", */
/* 				anObservation.event_id, */
/* 				anObservation.evlat, anObservation.evlon, anObservation.stlat, */
/* 				anObservation.stlon, anObservation.phs_velocity, observed, predicted, res, vel_res); */

		fprintf(fd2,"%s %10.4f %10.4f %10.4f %10.4f %10.4f %6.2f %10.3f %d\n",
			anObservation.event_id,
			anObservation.evlat, anObservation.evlon, anObservation.stlat,
			anObservation.stlon, anObservation.phs_velocity,anObservation.period,suggestedWeight,anObservation.iperiod);

		ires_bin = (res - min_res_bin) / res_bin_width;

		if (ires_bin < 0 || ires_bin >= nres_bins)
		{
			printf("Warning - residual out of bin range.\n");
			printf("%s %10.4f %10.4f %10.4f %10.4f %10.4f %6.2f %10.3f %10.3f\n",
				anObservation.event_id,
				anObservation.evlat, anObservation.evlon, anObservation.stlat,
				anObservation.stlon, anObservation.phs_velocity,anObservation.period,suggestedWeight, res);
		}
		else
			resbins[ires_bin] += 1;

		//numRows += 1;
		sum += res * res;

		temp = res * anObservation.weight;
		temp *= temp;
		finalPeriodMisfits[anObservation.iperiod] += temp;

		vel_sum += vel_res * vel_res;
		theResiduals[numRows] = res;
		numRows += 1;
	}

	fclose(intfile);
	fclose(flofile);
	fclose(fd0);
	fclose(fd1);
	fclose(fd2);
	getStats(theResiduals, numRows, &theFinalMean, &theFinalVariance);

	/*********************************************************************************/
	/* Write the misfits for each period into period_misfits.txt  */
	/*********************************************************************************/
	fd2 = fopen("period_misfits.txt","w");
	fprintf(fd2,"Misfit is the square of the difference.\n");
	fprintf(fd2,"Iperiod Period Initial_Misfit Final_Misfit Variance_Reduction, NObs MeanV VarV\n");
	for(i=0;i<nImages;i++){
	
/*
meanvel=0.0 ; roughvel = 0.0;
*/
	  GetModelRough(i,s,nx,ny,&meanvel, &roughvel);
	  fprintf(fd2,"%3.3d %8.2f %10.3f %10.3f %6.1f %7d %8.3f %f\n",i,Period[i],
		  initialPeriodMisfits[i]/sumWtsPerPeriod[i],
		  finalPeriodMisfits[i]/sumWtsPerPeriod[i],
		  100*(initialPeriodMisfits[i]-finalPeriodMisfits[i])/initialPeriodMisfits[i],
		  nObsPerPeriod[i],meanvel,roughvel);
	}

	fclose(fd2);
	/*********************************************************************************/
	/* Write therms values into the log and onto stdout  */
	/*********************************************************************************/

	rms = sqrt(sum / (float) numRows);
	vel_rms = sqrt(vel_sum / (float) numRows);

	fprintf(stdout,"-------------------------------------------------------------------\n");
	fprintf(stdout, "\nDamper = %8.3e Smoother = %8.3e Aprior = %8.3e\nrms = %8.3f vel_rms = %8.3f\n", 
		damper, smoother, apmodelwt, rms, vel_rms);
	fprintf(stdout,"-------------------------------------------------------------------\n");
	fprintf(logFile,"-------------------------------------------------------------------\n");
	fprintf(logFile, "\nDamper = %8.3e Smoother = %8.3e Aprior = %8.3e\nrms = %8.3f vel_rms = %8.3f\n", 
		damper, smoother, apmodelwt, rms, vel_rms);
	fprintf(logFile,"-------------------------------------------------------------------\n");

	for(k=0;k<nImages;k++){

	  
	  fprintf(stdout, "Image: %03d\n",k);
	  fprintf(logFile,"Image: %03d\n",k);
	  
	  fprintf(stdout, "Used A Least Squares Optimal Velocity (1/slowness): %.3f\n",dc_velocity[k]);
	  fprintf(logFile,"Used A Least Squares Optimal Velocity (1/slowness): %.3f\n",dc_velocity[k]);
	  fprintf(stdout, "Slowness model length (per cell): %.3f\n",modelLength[k]);
	  fprintf(logFile, "Slowness model length (per cell): %.3f\n",modelLength[k]);
	  fprintf(stdout,"Slowness model roughness (per cell): %.3f\n",modelRoughness[k]);
	  fprintf(logFile,"Slowness model roughness (per cell): %.3f\n",modelRoughness[k]);	
	  fprintf(logFile,"-------------------------------------------------------------------\n");
	  fprintf(stdout ,"-------------------------------------------------------------------\n");
	}

	fprintf(stdout, "Initial RMS = %10.3f Velocity RMS = %10.3f\n", initialRMS, initialVelocityRMS);
	fprintf(logFile,"Initial RMS = %10.3f Velocity RMS = %10.3f\n", initialRMS, initialVelocityRMS);
	fprintf(stdout, "Final   RMS = %10.3f Velocity RMS = %10.3f\n", rms, vel_rms);
	fprintf(logFile,"Final   RMS = %10.3f Velocity RMS = %10.3f\n", rms, vel_rms);
	fprintf(stdout, "Travel Time Variance Reduction: %6.1f\n", 100-100*rms/initialRMS);
	fprintf(logFile,"Travel Time Variance Reduction: %6.1f\n", 100-100*rms/initialRMS);
	fprintf(stdout, "   Velocity Variance Reduction: %6.1f\n", 100-100*vel_rms/initialVelocityRMS);
	fprintf(logFile,"   Velocity Variance Reduction: %6.1f\n", 100-100*vel_rms/initialVelocityRMS);
	fprintf(stdout, "Initial Residual Mean = %10.3f Standard Deviation = %10.3f\n",theInitialMean, sqrt(theInitialVariance));
	fprintf(logFile,"Initial Residual Mean = %10.3f Standard Deviation = %10.3f\n",theInitialMean, sqrt(theInitialVariance));
	fprintf(stdout, "Final Residual Mean   = %10.3f Standard Deviation = %10.3f\n",  theFinalMean, sqrt(theFinalVariance));
	fprintf(logFile,"Final Residual Mean   = %10.3f Standard Deviation = %10.3f\n",  theFinalMean, sqrt(theFinalVariance));
	fprintf(logFile,"-------------------------------------------------------------------\n");
	fprintf(stdout, "-------------------------------------------------------------------\n");

	/*********************************************************************************/
	/* Write the velocities into the file "velocities"  */
	/*********************************************************************************/
	for (i = 0; i < nCells; i++) {
	  if (fabs(s[i]) > kSmall)
	    v[i] = (1 / s[i]);
	  else
	    v[i] = 0;
	}


	if (verbose)
		fprintf(stdout, "Outputting the final model.\n");

	fd = creat("velocities", 0664);
	n = write(fd, (char *) v,nCells * sizeof(float));
	fd = creat("slownesses", 0664);
	n = write(fd, (char *) s, nCells * sizeof(float));

	/*********************************************************************************/
	/* Output the binned final residuals  */
	/*********************************************************************************/

	rcfd = fopen("resbins_final", "w");
	for (i = 0; i < nres_bins; i++) {
		fprintf(rcfd, "%f %d\n", min_res_bin + (i + 0.5) * res_bin_width, resbins[i]);
	}
	fclose(rcfd);

	fclose(logFile);

	return (0);
}
/* end of main */

/*****************************************************************************************/
void getStats(float *x, int n, float *mean, float *variance)
{
  int i;
  float sum;

  sum = 0.0;
  
  for(i=0;i<n;i++)
    {
      sum += x[i];
    }

  *mean = sum / (float) n;

  sum = 0.0;

  for(i=0;i<n;i++)
    {

      sum += (x[i] - *mean)*(x[i] - *mean);
    }

  *variance = sum / (float) (n-1);

}
